@@ -1,13 +1,13 @@
 /************************************************************************
 *                                                                       *
 *             Star Traders: A Game of Interstellar Trading              *
-*                Copyright (C) 1990-2021, John Zaitseff                 *
+*                Copyright (C) 1990-2022, John Zaitseff                 *
 *                                                                       *
 ************************************************************************/
 
 /*
   Author: John Zaitseff <J.Zaitseff@zap.org.au>
-  $Id: b56b67230a9acf37651ddb9d4a10204c87d30200 $
+  $Id: 9df1c185ddca553fbd5c0b3eecef692dd3b89fb3 $
 
   This file, fileio.c, contains the implementation of the game load and
   save functions used in Star Traders.
@@ -246,7 +246,7 @@ bool load_game (int num)
     filename = game_filename(num);
     assert(filename != NULL);
 
-    file = fopen(filename, "r");
+    file = game_fopen(num, "r");
     if (file == NULL) {
 	// File could not be opened
 
@@ -438,6 +438,7 @@ bool load_game (int num)
 bool save_game (int num)
 {
     const char *data_dir;
+    int data_dir_fd;
     char *buf, *encbuf;
     char *filename;
     FILE *file;
@@ -464,7 +465,8 @@ bool save_game (int num)
 
     // Create the data directory, if needed
     data_dir = data_directory();
-    if (data_dir != NULL) {
+    data_dir_fd = data_directory_fileno();
+    if (data_dir != NULL && data_dir_fd == -1) {
 	if (xmkdir(data_dir, S_IRWXU | S_IRWXG | S_IRWXO) != 0) {
 	    // Data directory could not be created
 	    saved_errno = errno;
@@ -484,7 +486,7 @@ bool save_game (int num)
     filename = game_filename(num);
     assert(filename != NULL);
 
-    file = fopen(filename, "w");
+    file = game_fopen(num, "w");
     if (file == NULL) {
 	// File could not be opened for writing
 	saved_errno = errno;
