@@ -1,13 +1,13 @@
 /************************************************************************
 *                                                                       *
 *             Star Traders: A Game of Interstellar Trading              *
-*                Copyright (C) 1990-2021, John Zaitseff                 *
+*                Copyright (C) 1990-2022, John Zaitseff                 *
 *                                                                       *
 ************************************************************************/
 
 /*
   Author: John Zaitseff <J.Zaitseff@zap.org.au>
-  $Id: 2fd36c7e95df9c18ee119f1b158de8718911a1fd $
+  $Id: 311e0a2e138e27febb9c9e403d09b532fb623f7b $
 
   This file, utils.c, contains the implementation of various utility
   functions used in Star Traders.
@@ -56,6 +56,7 @@ wchar_t *mon_thousands_sep;		// Locale's monetary thousands separator
 #define APPIMAGE_NAME		"AppRun.wrapped"
 
 #define DIRSEP			"/"		// Directory separator
+#define CURDIR			"."		// Current directory
 #define HIDDEN_PATH		"."		// Hidden file start char
 #define XDG_DATA_DEFAULT	".local" DIRSEP "share"
 
@@ -230,6 +231,7 @@ static const unsigned char xor_table[] = {
 
 static char *home_directory_str = NULL;		// Full pathname to home
 static char *data_directory_str = NULL;		// Writable data dir pathname
+static int data_directory_fd    = -1;		// Data dir file descriptor
 
 static bool is_posix_locale = false;		// Override strfmon()?
 
@@ -238,6 +240,31 @@ static bool is_posix_locale = false;		// Override strfmon()?
 *                  Module-specific function prototypes                  *
 ************************************************************************/
 
+/*
+  Function:   home_directory - Return home directory pathname
+  Parameters: (none)
+  Returns:    const char *   - Pointer to home directory
+
+  This function returns the full pathname to the user's home directory,
+  using the HOME environment variable.  Note that the existence of or
+  ability to write to this pathname is NOT checked by this function.
+  NULL is returned if the home directory cannot be determined.
+*/
+static const char *home_directory (void);
+
+
+/*
+  Function:   game_rel_filename - Generate a relative game filename
+  Parameters: gamenum           - Game number (1-9) as an integer
+  Returns:    char *            - Pointer to game filename string
+
+  This function returns a game filename, as a malloc()ed string, relative
+  to some current directory.  If gamenum is outside the range 1 to 9
+  (inclusive), NULL is returned.
+*/
+extern char *game_rel_filename (int gamenum);
+
+
 /*
   Function:   apply_xor - Scramble a buffer using xor_table
   Parameters: dest      - Location of destination buffer
@@ -301,7 +328,8 @@ static ssize_t b64decode (const void *restrict in, size_t inlen,
 *          Initialisation and environment function definitions          *
 ************************************************************************/
 
-// These functions are documented in the file "utils.h"
+/* These functions are documented in the file "utils.h" or in the
+   comments above. */
 
 
 /***********************************************************************/
@@ -348,6 +376,8 @@ const char *home_directory (void)
 
 	if (home != NULL && *home != '\0') {
 	    home_directory_str = xstrdup(home);
+	} else {
+	    home_directory_str = xstrdup(CURDIR);
 	}
     }
 
@@ -369,8 +399,6 @@ const char *data_directory (void)
     const char *dirsep_hiddenpath = DIRSEP HIDDEN_PATH;
     const char *datahome_part     = DIRSEP XDG_DATA_DEFAULT DIRSEP;
 
-    struct stat statbuf;
-
 
     assert(strlen(dirsep) == 1);
 
@@ -389,7 +417,8 @@ const char *data_directory (void)
 		strcat(p, dirsep_hiddenpath);
 		strcat(p, program_name);
 
-		if (stat(p, &statbuf) == 0 && S_ISDIR(statbuf.st_mode)) {
+		data_directory_fd = open(p, O_DIRECTORY | O_SEARCH);
+		if (data_directory_fd != -1) {
 		    // Directory "$HOME/.trader" exists: use that
 		    data_directory_str = p;
 		} else {
@@ -431,10 +460,48 @@ const char *data_directory (void)
 	}
     }
 
+    if (data_directory_str != NULL && data_directory_fd == -1) {
+	data_directory_fd = open(data_directory_str, O_DIRECTORY | O_SEARCH);
+    }
+
     return data_directory_str;
 }
 
 
+/***********************************************************************/
+// data_directory_fileno: Return file descriptor of data directory
+
+int data_directory_fileno (void)
+{
+    if (data_directory_fd == -1) {
+	(void) data_directory();
+    }
+
+    return data_directory_fd;
+}
+
+
+/***********************************************************************/
+// game_rel_filename: Generate a relative game filename
+
+char *game_rel_filename (int gamenum)
+{
+    /* This implementation assumes a POSIX environment and an ASCII-safe
+       character encoding. */
+
+    char *p;
+
+
+    if (gamenum < 1 || gamenum > 9) {
+	return NULL;
+    }
+
+    p = xmalloc(GAME_FILENAME_BUFSIZE);
+    snprintf(p, GAME_FILENAME_BUFSIZE, GAME_FILENAME_PROTO, gamenum);
+    return p;
+}
+
+
 /***********************************************************************/
 // game_filename: Convert an integer to a game filename
 
@@ -445,30 +512,70 @@ char *game_filename (int gamenum)
 
     const char *dirsep = DIRSEP;
 
-    char buf[GAME_FILENAME_BUFSIZE];	// Buffer for part of filename
+    char *gfn;				// Relative game filename
     const char *dd;			// Data directory
 
-
-    if (gamenum < 1 || gamenum > 9) {
+    gfn = game_rel_filename(gamenum);
+    if (gfn == NULL) {
 	return NULL;
     }
 
     dd = data_directory();
-    snprintf(buf, GAME_FILENAME_BUFSIZE, GAME_FILENAME_PROTO, gamenum);
-
     if (dd == NULL) {
-	return xstrdup(buf);
+	return gfn;
     } else {
-	char *p = xmalloc(strlen(dd) + strlen(dirsep) + strlen(buf) + 1);
+	char *p = xmalloc(strlen(dd) + strlen(dirsep) + strlen(gfn) + 1);
 
 	strcpy(p, dd);
 	strcat(p, dirsep);
-	strcat(p, buf);
+	strcat(p, gfn);
+	free(gfn);
 	return p;
     }
 }
 
 
+/***********************************************************************/
+// game_fopen: Open a game file for reading or writing
+
+FILE *game_fopen (int gamenum, const char *mode)
+{
+    char *gfn;
+    int ddfd;
+    int fd;
+    int oflag;
+    mode_t omode;
+
+
+    gfn = game_rel_filename(gamenum);
+    if (gfn == NULL) {
+	return NULL;
+    }
+
+    ddfd = data_directory_fileno();
+    if (ddfd == -1) {
+	ddfd = AT_FDCWD;
+    }
+
+    if (strcmp(mode, "r") == 0) {
+	oflag = O_RDONLY;
+	omode = 0;
+    } else if (strcmp(mode, "w") == 0) {
+	oflag = O_WRONLY | O_CREAT | O_TRUNC;
+	omode = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
+    } else {
+	free(gfn);
+	errno = EINVAL;
+	return NULL;
+    }
+
+    fd = openat(ddfd, gfn, oflag, omode);
+    free(gfn);
+
+    return (fd == -1) ? NULL : fdopen(fd, mode);
+}
+
+
 /************************************************************************
 *                 Error-reporting function definitions                  *
 ************************************************************************/
@@ -822,6 +929,7 @@ char *scramble (char *restrict dest, const char *restrict src,
 	// Place checksum in front of Base64 string
 	snprintf(chksumbuf, SCRAMBLE_CHKSUM_LEN + 1, "%03x", chksum);
 	memcpy(dest, chksumbuf, SCRAMBLE_CHKSUM_LEN);
+	dest[size - 1] = '\0';		// Keep Coverity Scan happy...
 
 	free(xorbuf);
     }
@@ -1116,19 +1224,21 @@ int xmkdir (const char *pathname, mode_t mode)
 	return -1;
     }
 
-    // Check that pathname already exists and is a directory
-    if (stat(pathname, &statbuf) == 0) {
-	if (S_ISDIR(statbuf.st_mode)) {
-	    return 0;
-	} else {
-	    errno = ENOTDIR;
-	    return -1;
-	}
-    }
-
     // Try creating the directory
     ret = mkdir(pathname, mode);
-    if (ret == 0 || (errno != ENOENT && errno != ENOTDIR)) {
+    if (ret != 0 && errno == EEXIST) {
+	// Pathname exists: is it a directory?
+	if (stat(pathname, &statbuf) == 0) {
+	    if (S_ISDIR(statbuf.st_mode)) {
+		return 0;
+	    } else {
+		errno = ENOTDIR;
+		return -1;
+	    }
+	} else {
+	    return -1;
+	}
+    } else if (ret == 0 || (errno != ENOENT && errno != ENOTDIR)) {
 	return ret;
     }
 
