@@ -1,13 +1,13 @@
 /************************************************************************
 *                                                                       *
 *             Star Traders: A Game of Interstellar Trading              *
-*                Copyright (C) 1990-2021, John Zaitseff                 *
+*                Copyright (C) 1990-2022, John Zaitseff                 *
 *                                                                       *
 ************************************************************************/
 
 /*
   Author: John Zaitseff <J.Zaitseff@zap.org.au>
-  $Id: 8482c90b988d300a78cf50e40413c1150eafffd4 $
+  $Id: 9f619302bc601b918ed8a60fbfa38186265f8299 $
 
   This file, utils.h, contains declarations for various utility functions
   used in Star Traders.
@@ -81,26 +81,14 @@ extern wchar_t *mon_thousands_sep;	// Local monetary thousands separator
 extern void init_program_name (const char *argv0);
 
 
-/*
-  Function:   home_directory - Return home directory pathname
-  Parameters: (none)
-  Returns:    const char *   - Pointer to home directory
-
-  This function returns the full pathname to the user's home directory,
-  using the HOME environment variable.  Note that the existence of or
-  ability to write to this pathname is NOT checked by this function.
-  NULL is returned if the home directory cannot be determined.
-*/
-extern const char *home_directory (void);
-
-
 /*
   Function:   data_directory - Return writable data directory pathname
   Parameters: (none)
   Returns:    const char *   - Pointer to data directory
 
   This function returns the full pathname to a potentially-writable data
-  directory, usually within the user's home directory.
+  directory, usually within the user's home directory.  The pointer so
+  returned must NOT be passed to free().
 
   Assuming program_name is set to "trader", if "$HOME/.trader" exists,
   that directory is returned as the data directory.  Otherwise, if the
@@ -109,12 +97,27 @@ extern const char *home_directory (void);
   "$HOME/.local/share/trader" is returned.
 
   Note that the returned path is NOT created by this function, nor is the
-  ability to read from or write to this path checked.  NULL is returned
-  if the path cannot be determined.
+  ability to read from or write to this path checked.  However, if the
+  path already exists, it is opened and used for game_fopen(); this may
+  be checked by running data_directory_fileno().  NULL is returned if the
+  path cannot be determined.
 */
 extern const char *data_directory (void);
 
 
+/*
+  Function:   data_directory_fileno - Return file descriptor of data directory
+  Parameters: (none)
+  Returns:    int                   - File descriptor, or -1 on failure
+
+  This function returns the file descriptor of the potentially-writable
+  data directory returned by data_directory(), or -1 if that directory
+  does not exist.  Note that this function does NOT create the data
+  directory itself.
+*/
+extern int data_directory_fileno (void);
+
+
 /*
   Function:   game_filename - Convert an integer to a game filename
   Parameters: gamenum       - Game number (1-9) as an integer
@@ -122,14 +125,32 @@ extern const char *data_directory (void);
 
   This function returns the full game filename as a malloc()ed string
   (ie, a string that must be freed at a later time by calling free()).
+  This string is to be used for display ONLY, not as a parameter to
+  fopen(), to avoid issues of TOCTOU (time-of-check / time-of-use).
 
   If gamenum is between 1 and 9 inclusive, the string returned is in the
-  form data_directory() + "/" + GAME_FILENAME(gamenum).  If gamenum is
-  any other integer, NULL is returned.
+  form DATA_DIRECTORY + "/" + GAME_FILENAME(gamenum).  If gamenum is any
+  other integer, NULL is returned.
 */
 extern char *game_filename (int gamenum);
 
 
+/*
+  Function:   game_fopen - Open a game file for reading or writing
+  Parameters: gamenum    - Game number (1-9) as an integer
+              mode       - File mode: "r" for reading, "w" for writing
+  Returns:    FILE *     - Pointer to open file structure
+
+  This function opens a game file for reading or writing, depending on
+  the mode string, returning a pointer to the open file structure.  Note
+  that the data directory must be created before calling this function,
+  by checking whether data_directory_fileno() returns -1: if so, the
+  directory must be created.  NULL is returned (and errno set) if the
+  file cannot be opened successfully.
+*/
+extern FILE *game_fopen (int gamenum, const char *mode);
+
+
 /************************************************************************
 *                  Error-reporting function prototypes                  *
 ************************************************************************/
